% to test the semi supervised ranking algorithm: multi-kernal ranking method 
% such as dhfr, dhfr and so on
% and study the test performance vs the number of training samples
% optimal parameters(delta, lambda) in CV process are used
% L=K^{-1} is used, it is equivalent to supervised ranking.
% dhfr cyclooxygenase-2 inhibitors øƼ
% in dhfr.mat in the first(set) column 1 represents training data, 2 for test, 3 for inactive compounds
% 2nd column activities is pIC50 value, namely, the real label
% 3-5 columns won't be used, the real feature dimension is 74

% first written by rushi lan, modified from graphrank_no_dhfr_lan.m and
% graphrank_no_dhfr.m
% Reference: Yicong Zhou, Chen Hong, Rushi Lan, and Zhibin Pan,
% "Generalization Performance of Regularized Ranking with Multiscale Kernels," IEEE Transactions on Neural Networks and Learning Systems, in press, 2015.
% rslan2012@gmail.com

clear
clc
load dhfr    % dhfr is with size 397*72
total_Xtrain = dhfr(dhfr(:,1)==1,3:end);  % 237*70
total_ytrain = dhfr(dhfr(:,1)==1,2);
Xtest = dhfr(dhfr(:,1)==2,3:end);   % 124*70
ytest = dhfr(dhfr(:,1)==2,2);
[total_train_num,fea_dim] = size(total_Xtrain);
test_num = size(Xtest,1);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% feature values are normalized to [0,1]
max_score = max([total_Xtrain;Xtest]);
min_score = min([total_Xtrain;Xtest]);
total_Xtrain = (total_Xtrain-ones(total_train_num,1)*min_score)./(ones(total_train_num,1)*(max_score-min_score));
Xtest = (Xtest-ones(test_num,1)*min_score)./(ones(test_num,1)*(max_score-min_score));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% to preassign the parameters
repeat_times = 30;       % repeated test times
sample_num = 10;     % number of samples in a portion
portion_num = floor(total_train_num/sample_num);
kernal_num=2;                 % number of used kernals, first try 2 
% delta = 4^1;         % the optimal width parameter of Gaussian kernel in CV process
delta=[4^2 64];      % the width parameter of each kernel; need to be optimal
v=[1 0.5];         % weight for each kernal; need to be optimal
lambda = 1e-3;       % the optimal regularization coefficient
%%%%%
train_time = zeros(repeat_times,portion_num);
test_time = zeros(repeat_times,portion_num);
%%%%%% performance measure in terms of ranking error, msd and tau
test_rwerr = zeros(repeat_times,portion_num);  % relevance-weighted ranking error for testing data
test_corr = zeros(repeat_times,portion_num);  % Pearson correlation coefficient
test_tau = zeros(repeat_times,portion_num);  % Kendall's tau coefficient
test_rho = zeros(repeat_times,portion_num);  % Spearman's rho coefficient
for ireptime = 1:repeat_times
    fprintf('In graphrank_no_dhfr,ireptime=%f,\n',ireptime)
    train_perm = randperm(total_train_num);
    for iportion = 1:portion_num
        % fprintf('In graphrank_no_dhfr,ireptime=%f,\n',ireptime)
        Xtrain = total_Xtrain(train_perm(1:sample_num*iportion),:);
        ytrain = total_ytrain(train_perm(1:sample_num*iportion),:);
        train_num = length(ytrain);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% compute the kernel matrix
        ttrain = cputime;
        Ktrain = zeros(train_num);             % Ktrain stands for K_{\bar{x}\bar{x}}
        for i=1:train_num
            for j=i:train_num
                for l=1:kernal_num
                squa_dist = sum((Xtrain(i,:)-Xtrain(j,:)).^2);
                Ktrain(i,j,l)=exp(-squa_dist/delta(l));
                end
            end
        end
        D = 2*train_num*eye(train_num)-2*ones(train_num);
        Y=[];
        temp=[];
        Ktrain_bar=zeros(train_num,train_num,l);
        for l=1:kernal_num
            Ktrain(:,:,l) = Ktrain(:,:,l)+Ktrain(:,:,l)'-diag(diag(Ktrain(:,:,l)));
            Ktrain_bar(:,:,l) = 2/train_num*D* Ktrain(:,:,l);
            Y=[Y;  2/train_num*D*ytrain];
            temp=[temp Ktrain_bar(:,:,l)];
        end
        A=repmat(temp,l,1);
        % [mm,nn]=size(eye(train_num));
        A2=A;
        for l=0:kernal_num-1
            temp=train_num*lambda*v(l+1)*eye(train_num);
            A2((1+l*train_num):(l+1)*train_num,(1+l*train_num):(l+1)*train_num)=A((1+l*train_num):(l+1)*train_num,(1+l*train_num):(l+1)*train_num)+temp;
        end
        % alpha = bicg(A2,Y);
        alpha = inv(A2)*Y;
        % alpha = bicg(A2'*A2,A2'*Y);
        alpha = alpha';
        % Ktrain = Ktrain+Ktrain'-diag(diag(Ktrain));
        train_time(ireptime,iportion) = cputime-ttrain;
        %%%%%%%%%%%%%%%%%%%%%%%% predict the test data  in cross validation process
        ttest = cputime;
        Kcross = zeros(train_num,test_num);
        for i = 1:train_num
            for j = 1:test_num
                for l=1:kernal_num
                squa_dist = sum((Xtrain(i,:)-Xtest(j,:)).^2);
                Kcross(i,j,l) = exp(-squa_dist/delta(l));
                end
            end
        end

        [sortedytest,idtest] = sort(ytest,'descend');
        alpha2=reshape(alpha,[train_num length(alpha)/train_num])';
        ytest_pred=ytest; 
        ytest_pred(:)=0;
        for l=1:kernal_num
             ytest_pred =  ytest_pred  + (alpha2(l,:)*Kcross(:,:,l))';      
        end
        % ytest_pred = (alpha*Kcross)';
        
        %%%%%%%%%%%%%%%%%%%%%%  compute the performance measures including Kendall's tau coefficient
        test_wrong_num = 0;     % the number of misranked pairs
        for i = 1:test_num-1
            for j = i+1:test_num
                pred_margin = ytest_pred(idtest(i))-ytest_pred(idtest(j));
                real_margin = ytest(idtest(i))-ytest(idtest(j));
                if  pred_margin < 0
                    test_wrong_num = test_wrong_num+1;
                    test_rwerr(ireptime,iportion) = test_rwerr(ireptime,iportion)+real_margin;
                elseif  pred_margin == 0    % in case of ties
                    test_wrong_num = test_wrong_num+0.5;
                    test_rwerr(ireptime,iportion) = test_rwerr(ireptime,iportion)+real_margin*0.5;
                end
            end
        end
        test_rwerr(ireptime,iportion) = 2*test_rwerr(ireptime,iportion)/(test_num*(test_num-1));
        test_corr(ireptime,iportion) = corr(ytest,ytest_pred,'type','pearson');
        test_tau(ireptime,iportion) = 1-4*sum(test_wrong_num(:))/(test_num*(test_num-1));
        test_rho(ireptime,iportion) = corr(ytest,ytest_pred,'type','spearman');
        test_time(ireptime,iportion) = cputime-ttest;
    end
end
test_rwerr_mean = mean(test_rwerr);  % mean percentage of pairwise misrankings for testing data
test_rwerr_std = std(test_rwerr);
test_corr_mean = mean(test_corr);  % mean test corr
test_corr_std = std(test_corr);
test_tau_mean = mean(test_tau);  % mean test tau
test_tau_std = std(test_tau);
test_rho_mean = mean(test_rho);  % mean test rho
test_rho_std = std(test_rho);
train_time_mean = mean(train_time);     train_time_std = std(train_time);
test_time_mean = mean(test_time);       test_time_std = std(test_time);
result = [test_rwerr_mean;test_rwerr_std;test_corr_mean;test_corr_std;test_tau_mean;test_tau_std;...
    test_rho_mean;test_rho_std;train_time_mean;train_time_std;test_time_mean;test_time_std];

figure, plot(test_corr_mean,'-d'); hold on
plot(test_tau_mean,'-rs');
plot(test_rho_mean,'-kv');
legend('error','\tau','\rho')

